#import <Foundation/Foundation.h>
#import "All.h"
#import "MACLib.h"
#import "aiff_process.h"
#import "CocoaMonkeyController.h"
#import "CMVorbisEncProtocol.h"
#import "CMFLACEncProtocol.h"
#import <sndfile.h>


#define SIZEOF_BUFFER 16384

struct soundInfo
{
	short channels;
	short bytesPerChannel;
	short bitsPerChannel;
	int samplingRate;
	unsigned int approxFileSize;
	unsigned int dataBytes;
};

int decodeToAIFF(char *inFile, char *outFile, id delegate, int *killFlag)
{
	int errnum;
	unsigned char *buffer,*buffer_converted;
	int received,nblock=0,blockper,i;
	unsigned int realbuffersize;
	struct soundInfo sinfo;
	SF_INFO sfinfo;
	SNDFILE *snd_data;
	FILE *fp;
	
	if((fp = fopen(outFile, "wb")) == NULL) {
		return ERROR_IO_WRITE;
	}
	fclose(fp);
	
	IAPEDecompress *decomp = CreateIAPEDecompress(inFile, &errnum);
	if(errnum != 0) {
		return errnum;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	sfinfo.samplerate = sinfo.samplingRate;
	sfinfo.channels = sinfo.channels;
	sfinfo.sections = 1;
	sfinfo.seekable = 1;
	sfinfo.frames = (sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels);
	switch(sinfo.bytesPerChannel) {
	  case 1:
		sfinfo.format = 0x20001;
		break;
	  case 2:
		sfinfo.format = 0x20002;
		break;
	  case 3:
		sfinfo.format = 0x20003;
		break;
	  default:
		delete decomp;
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	
	snd_data = sf_open(outFile, SFM_WRITE, &sfinfo);
	
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = (unsigned char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	if((buffer_converted = (unsigned char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	
	blockper = SIZEOF_BUFFER/(sinfo.channels);
	
	while(1) {
		if(*killFlag) {
			delete decomp;
			free(buffer);
			free(buffer_converted);
			sf_close(snd_data);
			return ERROR_USER_STOPPED_PROCESSING;
		}
		decomp->GetData((char *)buffer, blockper, &received);
		
		if(received<0) {
			delete decomp;
			free(buffer);
			free(buffer_converted);
			sf_close(snd_data);
			return ERROR_DECOMPRESSING_FRAME;
		}
		switch(sinfo.bytesPerChannel) {
		  case 1:
			for(i=0;i<received*sinfo.channels;i++) {
				if(*(buffer+i) <= 127)
					*(buffer_converted+i) = *(buffer+i) + 128;
				else
					*(buffer_converted+i) = *(buffer+i) - 128;
			}
			sf_write_raw(snd_data, buffer_converted, received*sinfo.channels);
			if(sf_error(snd_data)) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_WRITE;
			}
			break;
		  case 2:
			sf_write_short(snd_data, (short *)buffer, received*(sinfo.channels));
			if(sf_error(snd_data)) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_WRITE;
			}
			break;
		  case 3:
#ifdef _BIG_ENDIAN
			sf_write_raw(snd_data, buffer, received*(sinfo.channels)*3);
#else
			for(i=0;i<received*3*(sinfo.channels);i=i+3) {
				*(buffer_converted+i) = *(buffer+i+2);
				*(buffer_converted+i+1) = *(buffer+i+1);
				*(buffer_converted+i+2) = *(buffer+i);
			}
			sf_write_raw(snd_data, buffer_converted, received*(sinfo.channels)*3);
#endif
			if(sf_error(snd_data)) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_WRITE;
			}
			break;
		}
		nblock = nblock + received;
		[delegate setProgressBar:(int)(100.0*nblock/((sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels)))];
		if(received<blockper) break;
	}
	
	delete decomp;
	free(buffer);
	free(buffer_converted);
	sf_close(snd_data);
	
	return 0;

}

int decodeToPCM(char *inFile, char *outFile, id delegate, int *killFlag, int endian)
{
	int errnum;
	unsigned char *buffer, *buffer_converted;
	int received,nblock=0,blockper,i;
	unsigned int realbuffersize;
	struct soundInfo sinfo;
	FILE *fp;
	
	if((fp = fopen(outFile, "wb")) == NULL) {
		return ERROR_IO_WRITE;
	}
	
	IAPEDecompress *decomp = CreateIAPEDecompress(inFile, &errnum);
	if(errnum != 0) {
		return errnum;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = (unsigned char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	if((buffer_converted = (unsigned char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	
	blockper = SIZEOF_BUFFER/(sinfo.channels);
	
	while(1) {
		if(*killFlag) {
			delete decomp;
			free(buffer);
			return ERROR_USER_STOPPED_PROCESSING;
		}
#ifdef _BIG_ENDIAN
		if(endian == 2 && sinfo.bytesPerChannel != 1) decomp->GetData((char *)buffer, blockper, &received);
#else
		if(endian == 3 && sinfo.bytesPerChannel != 1) decomp->GetData((char *)buffer, blockper, &received);
#endif
		else decomp->GetData((char *)buffer_converted, blockper, &received);
		
		if(received<0) {
			delete decomp;
			free(buffer);
			free(buffer_converted);
			return ERROR_DECOMPRESSING_FRAME;
		}
		switch(sinfo.bytesPerChannel) {
		  case 1:
			/*
			for(i=0;i<received*sinfo.channels;i++) {
				if(*(buffer+i) <= 127)
					*(buffer_converted+i) = *(buffer+i) + 128;
				else
					*(buffer_converted+i) = *(buffer+i) - 128;
			}
			*/
			if(fwrite(buffer_converted, 1, received*sinfo.channels,fp) != received*sinfo.channels) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				return ERROR_IO_WRITE;
			}
			break;
		  case 2:
#ifdef _BIG_ENDIAN
			if(endian == 2) {
				for(i=0;i<received*2*(sinfo.channels);i=i+2) {
					*(buffer_converted+i) = *(buffer+i+1);
					*(buffer_converted+i+1) = *(buffer+i);
				}
			}
#else
			if(endian == 3) {
				for(i=0;i<received*2*(sinfo.channels);i=i+2) {
					*(buffer_converted+i) = *(buffer+i+1);
					*(buffer_converted+i+1) = *(buffer+i);
				}
			}
#endif
			if(fwrite(buffer_converted, 1, received*sinfo.channels*2,fp) != received*sinfo.channels*2) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				return ERROR_IO_WRITE;
			}
			break;
		  case 3:
#ifdef _BIG_ENDIAN
			if(endian == 2) {
				for(i=0;i<received*3*(sinfo.channels);i=i+3) {
					*(buffer_converted+i) = *(buffer+i+2);
					*(buffer_converted+i+1) = *(buffer+i+1);
					*(buffer_converted+i+2) = *(buffer+i);
				}
			}
#else
			if(endian == 3) {
				for(i=0;i<received*3*(sinfo.channels);i=i+3) {
					*(buffer_converted+i) = *(buffer+i+2);
					*(buffer_converted+i+1) = *(buffer+i+1);
					*(buffer_converted+i+2) = *(buffer+i);
				}
			}
#endif
			if(fwrite(buffer_converted, 1, received*sinfo.channels*3,fp) != received*sinfo.channels*3) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				return ERROR_IO_WRITE;
			}
			break;
		}
		nblock = nblock + received;
		[delegate setProgressBar:(int)(100.0*nblock/((sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels)))];
		if(received<blockper) break;
	}
	
	delete decomp;
	free(buffer);
	free(buffer_converted);
	fclose(fp);
	
	return 0;

}

int decodeToVorbis(char *inFile, char *outFile, id delegate, int *killFlag, float quality)
{
	int errnum;
	char *buffer;
	int received,nblock=0,blockper,ret;
	unsigned int realbuffersize;
	struct soundInfo sinfo;
	FILE *fp;
	if((fp = fopen(outFile, "wb")) == NULL) {
		return ERROR_IO_WRITE;
	}
	fclose(fp);
	
	IAPEDecompress *decomp = CreateIAPEDecompress(inFile, &errnum);
	if(errnum != 0) {
		return errnum;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	if(sinfo.bytesPerChannel != 2) {
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	
	
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = (char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	
	blockper = SIZEOF_BUFFER/(sinfo.channels);
	
	/* vorbis setup */
	
	Class VorbisEncoder = [delegate vorbisEncoderClass];
	
	id <CMVorbisEncProtocol>vorbisEncoder = [[VorbisEncoder alloc] initWithFile:outFile];
	if(vorbisEncoder == nil) {
		free(buffer);
		delete decomp;
		return ERROR_INITIALIZING_UNMAC;
	}
	ret = [vorbisEncoder setChannel: sinfo.channels frequency:sinfo.samplingRate bytes: sinfo.bytesPerChannel quality:quality];
	if(ret){
		free(buffer);
		delete decomp;
		[vorbisEncoder release];
		return ERROR_INITIALIZING_UNMAC;
	}
	
	[vorbisEncoder setBufferSize: blockper];
	
	/* start decoding */
	
	while(1) {
		if(*killFlag) {
			delete decomp;
			free(buffer);
			[vorbisEncoder release];
			return ERROR_USER_STOPPED_PROCESSING;
		}
		decomp->GetData(buffer, blockper, &received);
		
		if(received<0) {
			delete decomp;
			free(buffer);
			[vorbisEncoder release];
			return ERROR_DECOMPRESSING_FRAME;
		}
		if(received > 0) {
			[vorbisEncoder encodeBuffer: buffer blockNum: received];
		}
		
		nblock = nblock + received;
		[delegate setProgressBar:(int)(100.0*nblock/((sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels)))];
		if(received<blockper) break;
	}
	
	[vorbisEncoder finish];
	[vorbisEncoder release];
	delete decomp;
	free(buffer);
	return 0;
}

int decodeToFLAC(char *inFile, char *outFile, id delegate, int *killFlag, int compression)
{
	int errnum;
	char *buffer;
	int received,nblock=0,blockper,ret;
	unsigned int realbuffersize;
	struct soundInfo sinfo;
	FILE *fp;
	if((fp = fopen(outFile, "wb")) == NULL) {
		return ERROR_IO_WRITE;
	}
	fclose(fp);
	
	IAPEDecompress *decomp = CreateIAPEDecompress(inFile, &errnum);
	if(errnum != 0) {
		return errnum;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	/*
	if(sinfo.bytesPerChannel != 2) {
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	*/
	
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = (char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	
	
	blockper = SIZEOF_BUFFER/(sinfo.channels);
	
	/* FLAC setup */
	
	
	Class FLACEncoder = [delegate FLACEncoderClass];
	id <CMFLACEncProtocol>flacEncoder = [[FLACEncoder alloc] initWithFile:outFile];
	if(flacEncoder == nil) {
		free(buffer);
		delete decomp;
		return ERROR_INITIALIZING_UNMAC;
	}
	ret = [flacEncoder setChannel: sinfo.channels frequency:sinfo.samplingRate bytes: sinfo.bytesPerChannel dataSize: sinfo.dataBytes compressionLevel:compression];
	if(ret) {
		free(buffer);
		[flacEncoder release];
		delete decomp;
		return ERROR_INITIALIZING_UNMAC;
	}
	ret = [flacEncoder setBufferSize: realbuffersize];
	if(ret) {
		free(buffer);
		[flacEncoder release];
		delete decomp;
		return ERROR_INITIALIZING_UNMAC;
	}
	
	/* start decoding */
	
	while(1) {
		if(*killFlag) {
			delete decomp;
			free(buffer);
			[flacEncoder release];
			return ERROR_USER_STOPPED_PROCESSING;
		}
		decomp->GetData(buffer, blockper, &received);
		
		if(received<0) {
			delete decomp;
			free(buffer);
			[flacEncoder release];
			return ERROR_DECOMPRESSING_FRAME;
		}
		
		ret = [flacEncoder encodeBuffer: buffer blockNum: received];
		if(ret) {
			delete decomp;
			free(buffer);
			[flacEncoder release];
			return ERROR_DECOMPRESSING_FRAME;
		}
		
		nblock = nblock + received;
		[delegate setProgressBar:(int)(100.0*nblock/((sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels)))];
		if(received<blockper) break;
	}
	
	[flacEncoder finish];
	delete decomp;
	free(buffer);
	[flacEncoder release];
	return 0;
}

int encodeFromAIFF(char *inFile, char *outFile, int compressionLevel, id delegate, int *killFlag)
{
	int errnum;
	unsigned int i,realbuffersize, encoded=0, buffered;
	WAVEFORMATEX wavefmt;
	unsigned char *buffer, *buffer_converted;
	SF_INFO sfinfo;
	SNDFILE *snd_data;
	struct soundInfo sinfo;
	FILE *fp;
	
	if((fp = fopen(outFile, "wb")) == NULL) {
		return ERROR_IO_WRITE;
	}
	fclose(fp);
	
	snd_data = sf_open(inFile, SFM_READ, &sfinfo);
	sinfo.channels = sfinfo.channels;
	sinfo.samplingRate = sfinfo.samplerate;
	
	switch((sfinfo.format)&SF_FORMAT_SUBMASK) {
	  case 1:
		sinfo.bytesPerChannel = 1;
		sinfo.bitsPerChannel = 8;
		break;
	  case 2:
		sinfo.bytesPerChannel = 2;
		sinfo.bitsPerChannel = 16;
		break;
	  case 3:
		sinfo.bytesPerChannel = 3;
		sinfo.bitsPerChannel = 24;
		break;
	  case 5:
		sinfo.bytesPerChannel = 1;
		sinfo.bitsPerChannel = 8;
		break;
	  default:
		sf_close(snd_data);
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = (unsigned char *)malloc(realbuffersize))==NULL){
		sf_close(snd_data);
		return 0;
	}
	if((buffer_converted = (unsigned char *)malloc(realbuffersize))==NULL){
		sf_close(snd_data);
		return 0;
	}
	
	
	IAPECompress *comp = CreateIAPECompress(&errnum);
	if(errnum != 0) {
		free(buffer);
		free(buffer_converted);
		sf_close(snd_data);
		return errnum;
	}
	
	NSFileManager *fm = [NSFileManager defaultManager];
	sinfo.approxFileSize = [[[fm fileAttributesAtPath:[NSString stringWithUTF8String:inFile] traverseLink:YES] objectForKey:NSFileSize] unsignedIntValue];
	
	FillWaveFormatEx(&wavefmt, sinfo.samplingRate, sinfo.bitsPerChannel, sinfo.channels);
	
	comp->Start(outFile, &wavefmt, sinfo.approxFileSize, compressionLevel, NULL, CREATE_WAV_HEADER_ON_DECOMPRESSION);
	if(errnum != 0) {
		delete comp;
		free(buffer);
		free(buffer_converted);
		sf_close(snd_data);
		return errnum;
	}
	
	while(1) {
		if(*killFlag) {
			delete comp;
			free(buffer);
			free(buffer_converted);
			sf_close(snd_data);
			return ERROR_USER_STOPPED_PROCESSING;
		}
		switch(sinfo.bytesPerChannel) {
		  case 1: // 8bit
			buffered = sf_read_raw(snd_data, buffer, realbuffersize);
			if(sf_error(snd_data)) {
				delete comp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_READ;
			}
			if(((sfinfo.format)&SF_FORMAT_SUBMASK) == 1) {
				for(i=0;i<buffered;i++) {
					if(*(buffer+i) <= 127)
						*(buffer_converted+i) = *(buffer+i) + 128;
					else
						*(buffer_converted+i) = *(buffer+i) - 128;
				}
			}
			break;
		  case 2: // 16bit
			buffered = sf_read_short(snd_data, (short *)buffer_converted, realbuffersize/2);
			if(sf_error(snd_data)) {
				delete comp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_READ;
			}
			buffered = buffered * 2;
			break;
		  case 3: // 24bit
			sfinfo.format = sfinfo.format | 0x10000000;
#ifdef _BIG_ENDIAN
			buffered = sf_read_raw(snd_data, buffer_converted, realbuffersize);
#else
			buffered = sf_read_raw(snd_data, buffer, realbuffersize);
#endif
			if(sf_error(snd_data)) {
				delete comp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return ERROR_IO_READ;
			}
#ifndef _BIG_ENDIAN
			for(i=0;i<buffered;i=i+3) {
				*(buffer_converted+i) = *(buffer+i+2);
				*(buffer_converted+i+1) = *(buffer+i+1);
				*(buffer_converted+i+2) = *(buffer+i);
			}
#endif
			break;
		}
		if(buffered) {
			comp->AddData(buffer_converted, buffered);
			if(errnum != 0) {
				delete comp;
				free(buffer);
				free(buffer_converted);
				sf_close(snd_data);
				return errnum;
			}
			encoded = encoded + buffered;
			[delegate setProgressBar:(int)(100.0*encoded/(sinfo.approxFileSize))];
		}
		if(buffered < realbuffersize) {
			comp->Finish(NULL,0,0);
			break;
		}
	}
	
	delete comp;
	sf_close(snd_data);
	free(buffer);
	free(buffer_converted);
	
	return 0;
	
}

int decodeApeWithCue(char *inFile, char *outPath, id delegate, int *killFlag, int format, int renameDup)
{
	int errnum,stat=1,nindex=0,ntrack=1,type[99],size[99],lastFile,lastDecode,blank,block,len;
	unsigned char *buffer,*buffer_converted;
	char cuebuf[512];
	int received,nbytes=0,blockper,i,k,n,blockToReceive;
	unsigned int realbuffersize;
	struct soundInfo sinfo;
	SF_INFO sfinfo;
	SNDFILE *snd_data;
	FILE *fp;
	NSString *extension, *outFile;
	NSFileManager *fm = [NSFileManager defaultManager];
	
	if((fp = fopen(inFile, "rb")) == NULL) {
		return ERROR_IO_READ;
	}
	
	NSString *apeFile = [[[NSString stringWithUTF8String: inFile] stringByDeletingPathExtension] stringByAppendingPathExtension:@"ape"];
	NSString *outDir = [[NSString stringWithUTF8String: outPath] stringByAppendingPathComponent: [[apeFile lastPathComponent] stringByDeletingPathExtension]];
	mkdir([outDir UTF8String],0755);
	if(format == 1) extension = @"aif";
	else if(format == 0) extension = @"wav";
	else extension = @"pcm";
	
	IAPEDecompress *decomp = CreateIAPEDecompress([apeFile UTF8String], &errnum);
	if(errnum != 0) {
		return errnum;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	while(fgets(cuebuf,512,fp) != NULL) {
		len = strlen(cuebuf);
		if(stat == 1) {
			i=0;
			while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
			if(i>len-4) continue;
			if(!strncmp(cuebuf+i,"TRACK",5)) {
				stat = 2;
			}
		}
		else if(stat == 2) {
			i=0;
			while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
			if(i>len-4) continue;
			if(!strncmp(cuebuf+i,"INDEX",5)) {
				i = i + 5;
				while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
				if(!strncmp(cuebuf+i,"00",2)) {
					type[nindex] = 0;
					i = i + 2;
					while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
					*(cuebuf+i+2) = '\0';
					*(cuebuf+i+5) = '\0';
					*(cuebuf+i+8) = '\0';
					block = atoi(cuebuf+i)*60*sinfo.samplingRate;
					block = block + atoi(cuebuf+i+3)*sinfo.samplingRate;
					block = block + atoi(cuebuf+i+6)*sinfo.samplingRate/75;
					size[nindex++] = block;
				}
				else if(!strncmp(cuebuf+i,"01",2)) {
					type[nindex] = 1;
					i = i + 2;
					while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
					*(cuebuf+i+2) = '\0';
					*(cuebuf+i+5) = '\0';
					*(cuebuf+i+8) = '\0';
					block = atoi(cuebuf+i)*60*sinfo.samplingRate;
					block = block + atoi(cuebuf+i+3)*sinfo.samplingRate;
					block = block + atoi(cuebuf+i+6)*sinfo.samplingRate/75;
					stat = 1;
					size[nindex++] = block;
				}
			}
		}
	}
	fclose(fp);
	
	for(i=nindex-1;i>0;i--) {
		size[i] = size[i] - size[i-1];
	}
	
	sfinfo.samplerate = sinfo.samplingRate;
	sfinfo.channels = sinfo.channels;
	sfinfo.sections = 1;
	sfinfo.seekable = 1;
	sfinfo.frames = (sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels);
	switch(sinfo.bytesPerChannel) {
	  case 1:
		if(format == 1) sfinfo.format = 0x20001;
		else if(format == 2) sfinfo.format = 0x10040005;
		else if(format == 3) sfinfo.format = 0x20040005;
		else sfinfo.format = 0x10005;
		break;
	  case 2:
		if(format == 1) sfinfo.format = 0x20002;
		else if(format == 2) sfinfo.format = 0x10040002;
		else if(format == 3) sfinfo.format = 0x20040002;
		else sfinfo.format = 0x10002;
		break;
	  case 3:
		if(format == 1) sfinfo.format = 0x20003;
		else if(format == 2) sfinfo.format = 0x10040003;
		else if(format == 3) sfinfo.format = 0x20040003;
		else sfinfo.format = 0x10003;
		break;
	  default:
		delete decomp;
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = (unsigned char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	if((buffer_converted = (unsigned char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	
	for(n=0;n<nindex;n++) {
		k=1;
		blockper = SIZEOF_BUFFER/(sinfo.channels);
		if(n == nindex-1) {
			lastFile = 1;
		}
		else {
			lastFile = 0;
			blockToReceive = size[n+1];
		}
		if(type[n] == 0) {
			blank = 1;
		}
		else {
			outFile = [outDir stringByAppendingPathComponent:[NSString stringWithFormat:@"Track%02d.%@",ntrack++,extension]];
			if(renameDup) {
				if([fm fileExistsAtPath:outFile]) {
				outFile = [[outFile stringByDeletingPathExtension] stringByAppendingPathExtension:[NSString stringWithFormat:@"%d.%@",k,extension]];
					while([fm fileExistsAtPath:outFile]) {
						k++;
						outFile = [outFile stringByDeletingPathExtension];
						outFile = [[outFile stringByDeletingPathExtension] stringByAppendingPathExtension:[NSString stringWithFormat:@"%d.%@",k,extension]];
					}
				}
			}
			snd_data = sf_open([outFile UTF8String], SFM_WRITE, &sfinfo);
			if(sf_error(snd_data)) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				return ERROR_IO_WRITE;
			}
			blank = 0;
		}
		lastDecode = 0;
		while(1) {
			if(*killFlag) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				if(!blank) sf_close(snd_data);
				return ERROR_USER_STOPPED_PROCESSING;
			}
			if(!lastFile && blockper >= blockToReceive) {
				blockper = blockToReceive;
				lastDecode = 1;
			}

			decomp->GetData((char *)buffer, blockper, &received);
		
			if(received<0) {
				delete decomp;
				free(buffer);
				free(buffer_converted);
				if(!blank) sf_close(snd_data);
				return ERROR_DECOMPRESSING_FRAME;
			}
			if(!blank) {
				switch(sinfo.bytesPerChannel) {
				  case 1:
					if(format == 1) {
						for(i=0;i<received*sinfo.channels;i++) {
							if(*(buffer+i) <= 127)
								*(buffer_converted+i) = *(buffer+i) + 128;
							else
								*(buffer_converted+i) = *(buffer+i) - 128;
						}
						sf_write_raw(snd_data, buffer_converted, received*sinfo.channels);
					}
					else sf_write_raw(snd_data, buffer, received*sinfo.channels);
					if(sf_error(snd_data)) {
						delete decomp;
						free(buffer);
						free(buffer_converted);
						sf_close(snd_data);
						return ERROR_IO_WRITE;
					}
					break;
				  case 2:
					sf_write_short(snd_data, (short *)buffer, received*(sinfo.channels));
					if(sf_error(snd_data)) {
						delete decomp;
						free(buffer);
						free(buffer_converted);
						sf_close(snd_data);
						return ERROR_IO_WRITE;
					}
					break;
				  case 3:
#ifdef _BIG_ENDIAN
					if(format == 0 || format == 2) {
						for(i=0;i<received*3*(sinfo.channels);i=i+3) {
							*(buffer_converted+i) = *(buffer+i+2);
							*(buffer_converted+i+1) = *(buffer+i+1);
							*(buffer_converted+i+2) = *(buffer+i);
						}
						sf_write_raw(snd_data, buffer_converted, received*(sinfo.channels)*3);
					}
#else
					if(format == 1 || format == 3) {
						for(i=0;i<received*3*(sinfo.channels);i=i+3) {
							*(buffer_converted+i) = *(buffer+i+2);
							*(buffer_converted+i+1) = *(buffer+i+1);
							*(buffer_converted+i+2) = *(buffer+i);
						}
						sf_write_raw(snd_data, buffer_converted, received*(sinfo.channels)*3);
					}
#endif
					else sf_write_raw(snd_data, buffer, received*(sinfo.channels)*3);
					if(sf_error(snd_data)) {
						delete decomp;
						free(buffer);
						free(buffer_converted);
						sf_close(snd_data);
						return ERROR_IO_WRITE;
					}
					break;
				}
			}
			if(!lastFile) blockToReceive = blockToReceive - received;
			nbytes = nbytes + received*sinfo.bytesPerChannel*sinfo.channels;
			[delegate setProgressBar:(int)(100.0*nbytes/sinfo.dataBytes)];
			if(lastDecode) break;
			if(received<blockper) break;
		}
		if(!blank) sf_close(snd_data);
	}
	
	delete decomp;
	free(buffer);
	free(buffer_converted);
	
	return 0;

}

int decodeApeWithCueAndCompress(char *inFile, char *outPath, id delegate, int *killFlag, int format, int renameDup, float compress)
{
	int errnum,stat=1,nindex=0,ntrack=1,type[99],size[99],lastFile,lastDecode,blank,block,len,ret;
	char *buffer;
	char cuebuf[512];
	int received,nbytes=0,blockper,i,k,n,blockToReceive;
	unsigned int realbuffersize;
	struct soundInfo sinfo;
	SF_INFO sfinfo;
	SNDFILE *snd_data;
	FILE *fp;
	NSString *extension, *outFile;
	NSFileManager *fm = [NSFileManager defaultManager];
	
	if((fp = fopen(inFile, "rb")) == NULL) {
		return ERROR_IO_READ;
	}
	
	NSString *apeFile = [[[NSString stringWithUTF8String: inFile] stringByDeletingPathExtension] stringByAppendingPathExtension:@"ape"];
	NSString *outDir = [[NSString stringWithUTF8String: outPath] stringByAppendingPathComponent: [[apeFile lastPathComponent] stringByDeletingPathExtension]];
	mkdir([outDir UTF8String],0755);
	if(format == 4) extension = @"ogg";
	else extension = @"flac";
	
	IAPEDecompress *decomp = CreateIAPEDecompress([apeFile UTF8String], &errnum);
	if(errnum != 0) {
		return errnum;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	while(fgets(cuebuf,512,fp) != NULL) {
		len = strlen(cuebuf);
		if(stat == 1) {
			i=0;
			while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
			if(i>len-4) continue;
			if(!strncmp(cuebuf+i,"TRACK",5)) {
				stat = 2;
			}
		}
		else if(stat == 2) {
			i=0;
			while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
			if(i>len-4) continue;
			if(!strncmp(cuebuf+i,"INDEX",5)) {
				i = i + 5;
				while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
				if(!strncmp(cuebuf+i,"00",2)) {
					type[nindex] = 0;
					i = i + 2;
					while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
					*(cuebuf+i+2) = '\0';
					*(cuebuf+i+5) = '\0';
					*(cuebuf+i+8) = '\0';
					block = atoi(cuebuf+i)*60*sinfo.samplingRate;
					block = block + atoi(cuebuf+i+3)*sinfo.samplingRate;
					block = block + atoi(cuebuf+i+6)*sinfo.samplingRate/75;
					size[nindex++] = block;
				}
				else if(!strncmp(cuebuf+i,"01",2)) {
					type[nindex] = 1;
					i = i + 2;
					while(*(cuebuf+i)==' ' || *(cuebuf+i)=='\t') i++;
					*(cuebuf+i+2) = '\0';
					*(cuebuf+i+5) = '\0';
					*(cuebuf+i+8) = '\0';
					block = atoi(cuebuf+i)*60*sinfo.samplingRate;
					block = block + atoi(cuebuf+i+3)*sinfo.samplingRate;
					block = block + atoi(cuebuf+i+6)*sinfo.samplingRate/75;
					stat = 1;
					size[nindex++] = block;
				}
			}
		}
	}
	fclose(fp);
	
	for(i=nindex-1;i>0;i--) {
		size[i] = size[i] - size[i-1];
	}
	
	sfinfo.samplerate = sinfo.samplingRate;
	sfinfo.channels = sinfo.channels;
	sfinfo.sections = 1;
	sfinfo.seekable = 1;
	sfinfo.frames = (sinfo.dataBytes)/(sinfo.bytesPerChannel)/(sinfo.channels);
	
	if(sinfo.bytesPerChannel != 2 && format == 4) {
		delete decomp;
		return ERROR_INPUT_FILE_UNSUPPORTED_BIT_DEPTH;
	}
	
	realbuffersize = SIZEOF_BUFFER*(sinfo.bytesPerChannel);
	
	if((buffer = (char *)malloc(realbuffersize))==NULL){
		return ERROR_INSUFFICIENT_MEMORY;
	}
	
	blockper = SIZEOF_BUFFER/(sinfo.channels);
	
	id <CMVorbisEncProtocol>vorbisEncoder;
	id <CMFLACEncProtocol>flacEncoder;
	
	for(n=0;n<nindex;n++) {
		k=1;
		if(n == nindex-1) {
			lastFile = 1;
		}
		else {
			lastFile = 0;
			blockToReceive = size[n+1];
		}
		if(type[n] == 0) {
			blank = 1;
		}
		else {
			outFile = [outDir stringByAppendingPathComponent:[NSString stringWithFormat:@"Track%02d.%@",ntrack++,extension]];
			if(renameDup) {
				if([fm fileExistsAtPath:outFile]) {
				outFile = [[outFile stringByDeletingPathExtension] stringByAppendingPathExtension:[NSString stringWithFormat:@"%d.%@",k,extension]];
					while([fm fileExistsAtPath:outFile]) {
						k++;
						outFile = [outFile stringByDeletingPathExtension];
						outFile = [[outFile stringByDeletingPathExtension] stringByAppendingPathExtension:[NSString stringWithFormat:@"%d.%@",k,extension]];
					}
				}
			}
			if((fp = fopen([outFile UTF8String], "wb")) == NULL) {
				return ERROR_IO_WRITE;
				free(buffer);
			}
			fclose(fp);
			blank = 0;
			
			/* setup vorbis or FLAC */
			
			if(format == 4) {
				Class VorbisEncoder = [delegate vorbisEncoderClass];
	
				vorbisEncoder = [[VorbisEncoder alloc] initWithFile:(char *)[outFile UTF8String]];
				if(vorbisEncoder == nil) {
					free(buffer);
					delete decomp;
					return ERROR_INITIALIZING_UNMAC;
				}
				ret = [vorbisEncoder setChannel: sinfo.channels frequency:sinfo.samplingRate bytes: sinfo.bytesPerChannel quality:compress];
				if(ret){
					free(buffer);
					delete decomp;
					[vorbisEncoder release];
					return ERROR_INITIALIZING_UNMAC;
				}
				[vorbisEncoder setBufferSize: blockper];
			}
			else {
				Class FLACEncoder = [delegate FLACEncoderClass];
				flacEncoder = [[FLACEncoder alloc] initWithFile:(char *)[outFile UTF8String]];
				if(flacEncoder == nil) {
					free(buffer);
					delete decomp;
					return ERROR_INITIALIZING_UNMAC;
				}
				ret = [flacEncoder setChannel: sinfo.channels frequency:sinfo.samplingRate bytes: sinfo.bytesPerChannel dataSize: sinfo.dataBytes compressionLevel:(int)compress];
				if(ret) {
					free(buffer);
					[flacEncoder release];
					delete decomp;
					return ERROR_INITIALIZING_UNMAC;
				}
				ret = [flacEncoder setBufferSize: realbuffersize];
				if(ret) {
					free(buffer);
					[flacEncoder release];
					delete decomp;
					return ERROR_INITIALIZING_UNMAC;
				}
			}
		}
		lastDecode = 0;
		while(1) {
			if(*killFlag) {
				delete decomp;
				free(buffer);
				if(format == 4) [vorbisEncoder release];
				else [flacEncoder release];
				return ERROR_USER_STOPPED_PROCESSING;
			}
			if(!lastFile && blockper >= blockToReceive) {
				blockper = blockToReceive;
				lastDecode = 1;
			}

			decomp->GetData(buffer, blockper, &received);
		
			if(received<0) {
				delete decomp;
				free(buffer);
				if(format == 4) [vorbisEncoder release];
				else [flacEncoder release];
				return ERROR_DECOMPRESSING_FRAME;
			}
			if(!blank) {
				if(format == 4) {
					if(received > 0) {
						[vorbisEncoder encodeBuffer: buffer blockNum: received];
					}
				}
				else {
					ret = [flacEncoder encodeBuffer: buffer blockNum: received];
					if(ret) {
						delete decomp;
						free(buffer);
						[flacEncoder release];
						return ERROR_DECOMPRESSING_FRAME;
					}
				}
			}
			if(!lastFile) blockToReceive = blockToReceive - received;
			nbytes = nbytes + received*sinfo.bytesPerChannel*sinfo.channels;
			[delegate setProgressBar:(int)(100.0*nbytes/sinfo.dataBytes)];
			if(lastDecode) break;
			if(received<blockper) break;
		}
		if(!blank) {
			if(format == 4) [vorbisEncoder release];
			else [flacEncoder release];
		}
	}
	
	delete decomp;
	free(buffer);
	
	return 0;

}