//
//  CMVorbisEnc.m
//  CocoaMonkey
//
//  Created by tmkk on Fri Jul 16 2004.
//  Copyright (c) 2004 tmkk. All rights reserved.
//

#import "CMVorbisEnc.h"
#import <time.h>

ogg_stream_state os;
ogg_page         og;
ogg_packet       op;
vorbis_info      vi;
vorbis_comment   vc;
vorbis_dsp_state vd;
vorbis_block     vb;

@implementation CMVorbisEnc

- (id)initWithFile:(char *)file
{
	[super init];
	if((fp = fopen(file, "wb")) == nil) {
		return nil;
	}
	eos = 0;
	return self;
}

- (int)setChannel: (int)ch frequency:(int)freq bytes: (int)n quality:(float)q
{
	vorbis_info_init(&vi);
	int ret = vorbis_encode_init_vbr(&vi,ch,freq,q);
	if(ret){
		fclose(fp);
		return 1;
	}
	
	channels = ch;
	bytesPerChannel = n;
	
	vorbis_comment_init(&vc);
	vorbis_comment_add_tag(&vc,"ENCODER","CocoaMonkey Vorbis output");
	vorbis_analysis_init(&vd,&vi);
	vorbis_block_init(&vd,&vb);
	srand(time(NULL));
	ogg_stream_init(&os,rand());
	
	ogg_packet header;
	ogg_packet header_comm;
	ogg_packet header_code;
	
	vorbis_analysis_headerout(&vd,&vc,&header,&header_comm,&header_code);
	ogg_stream_packetin(&os,&header);
	ogg_stream_packetin(&os,&header_comm);
	ogg_stream_packetin(&os,&header_code);
	
	while(1){
		int result=ogg_stream_flush(&os,&og);
		if(result==0) break;
		fwrite(og.header,1,og.header_len,fp);
		fwrite(og.body,1,og.body_len,fp);
	}
	
	return 0;
}

- (void)setBufferSize: (int)size
{
	blockper = size;
}

- (void)encodeBuffer: (char *)buffer blockNum:(int)nblock
{
	float **buffer_converted;
	int i;
	buffer_converted = vorbis_analysis_buffer(&vd,blockper);
	for(i=0;i<nblock;i++){
#ifdef _BIG_ENDIAN
		buffer_converted[0][i]=((buffer[i*(channels*bytesPerChannel)]<<8)|(0x00ff&(int)buffer[i*(channels*bytesPerChannel)+1]))/32768.f;
		if(channels == 2)
			buffer_converted[1][i]=((buffer[i*(channels*bytesPerChannel)+2]<<8)|(0x00ff&(int)buffer[i*(channels*bytesPerChannel)+3]))/32768.f;
#else
		buffer_converted[0][i]=((buffer[i*(channels*bytesPerChannel)+1]<<8)|(0x00ff&(int)buffer[i*(channels*bytesPerChannel)]))/32768.f;
		if(channels == 2)
			buffer_converted[1][i]=((buffer[i*(channels*bytesPerChannel)+3]<<8)|(0x00ff&(int)buffer[i*(channels*bytesPerChannel)+2]))/32768.f;
#endif		
	}
	vorbis_analysis_wrote(&vd,nblock);
			
	while(vorbis_analysis_blockout(&vd,&vb)==1){
		vorbis_analysis(&vb,NULL);
		vorbis_bitrate_addblock(&vb);
		while(vorbis_bitrate_flushpacket(&vd,&op)){
			ogg_stream_packetin(&os,&op);
			while(!eos){
				int result=ogg_stream_pageout(&os,&og);
				if(result==0)break;
				fwrite(og.header,1,og.header_len,fp);
				fwrite(og.body,1,og.body_len,fp);
				if(ogg_page_eos(&og))eos=1;
			}
		}
	}
}

- (void)finish
{
	vorbis_analysis_wrote(&vd,0);
	while(vorbis_analysis_blockout(&vd,&vb)==1){
		vorbis_analysis(&vb,NULL);
		vorbis_bitrate_addblock(&vb);
		while(vorbis_bitrate_flushpacket(&vd,&op)){
			ogg_stream_packetin(&os,&op);
			while(!eos){
				int result=ogg_stream_pageout(&os,&og);
				if(result==0)break;
				fwrite(og.header,1,og.header_len,fp);
				fwrite(og.body,1,og.body_len,fp);
				if(ogg_page_eos(&og))eos=1;
			}
		}
	}
}

- (void)dealloc
{
	
	ogg_stream_clear(&os);
	vorbis_block_clear(&vb);
	vorbis_dsp_clear(&vd);
	vorbis_comment_clear(&vc);
	vorbis_info_clear(&vi);
	fclose(fp);
	[super dealloc];
}

@end
